from pokerrl_env.datatypes import CALL, CHECK, FOLD, INT_POSITIONS_BY_NUM_PLAYERS, PREFLOP_ORDER_BY_NUM_PLAYERS, POSITION_TO_SEAT, PLAYERS_POSITIONS_DICT, BetLimits, GameTypes, Street
from pokerrl_env.utils import calculate_fixed_limit_mask, calculate_no_limit_betsize, calculate_pot_limit_betsize, calculate_pot_limit_mask, calculate_no_limit_mask, return_deck

class Config:
    def __init__(self, game_type=GameTypes.OMAHA_HI, num_players=2, bet_limit=BetLimits.POT_LIMIT,
                 betsizes=(1, 0.9, 0.75, 0.67, 0.5, 0.33, 0.25, 0.1),
                 blinds=(0.5,1), stack_sizes=100,is_server=False):
        assert num_players >= 2, "Number of players must be at least 2"
        assert bet_limit in [BetLimits.POT_LIMIT, BetLimits.NO_LIMIT, BetLimits.FIXED_LIMIT], "Bet limit must be one of Pot limit, No limit, or Fixed limit"
        assert len(betsizes) > 0, "Betsizes must be a non-empty tuple"
        assert len(blinds) == 2, "Blinds must be a tuple of length 2"
        assert stack_sizes > 0, "Stack sizes must be a positive integer"
        self.is_server = is_server
        self.game_type = game_type
        self.num_players = num_players
        self.bet_limit = bet_limit
        self.betsizes = betsizes
        self.blinds = blinds
        self.stack_sizes = stack_sizes
        self.action_strs = [FOLD, CHECK, CALL]
        self.action_type_to_int = {a: i for i, a in enumerate(self.action_strs, start=1)}
        self.action_betsize_to_int = {b: i for i, b in enumerate(betsizes, start=4)}
        self.action_to_int = self.action_type_to_int | self.action_betsize_to_int
        self.action_to_int['_'] = 0
        self.action_to_str = {v: k for k, v in self.action_to_int.items()}
        self.action_to_str[0] = "None"
        self.num_actions = len(betsizes) + len(self.action_strs)
        self.player_positions = INT_POSITIONS_BY_NUM_PLAYERS[num_players]
        self.player_state_mapping, self.player_state_shape,self.global_state_mapping,self.global_state_shape = return_mappings(game_type)
        if bet_limit == BetLimits.POT_LIMIT:
            self.return_action_mask = calculate_pot_limit_mask
            self.return_betsize = calculate_pot_limit_betsize
        elif bet_limit == BetLimits.NO_LIMIT:
            self.return_action_mask = calculate_no_limit_mask
            self.return_betsize = calculate_no_limit_betsize
        elif bet_limit == BetLimits.FIXED_LIMIT:
            raise NotImplementedError("Fixed limit not implemented yet")
            # self.return_action_mask = calculate_fixed_limit_mask
            # self.return_betsize = calculate_fixed_limit_betsize
        else:
            raise ValueError("Bet limit must be one of Pot limit, No limit, or Fixed limit")

    @staticmethod
    def convert_model_action_to_state(model_action):
        return model_action + 1
    

def return_mappings(game_type):
    """ In the future will return different mappings depending on the game type """
    if game_type == GameTypes.OMAHA_HI:
        player_state_mapping = {
            "hand_range": [0, 8],
            "board_range": [8, 18],
            "street": 18,
            "num_players": 19,
            "hero_position": 20,
            "hero_active": 21,
            "vil_1_active": 22,
            "vil_2_active": 23,
            "vil_3_active": 24,
            "vil_4_active": 25,
            "vil_5_active": 26,
            "vil_1_position": 27,
            "vil_2_position": 28,
            "vil_3_position": 29,
            "vil_4_position": 30,
            "vil_5_position": 31,
            "last_agro_amount": 32,
            "last_agro_action": 33,
            "last_agro_position": 34,
            "last_agro_bet_is_blind": 35,
            "hero_stack": 36,
            "vil_1_stack": 37,
            "vil_2_stack": 38,
            "vil_3_stack": 39,
            "vil_4_stack": 40,
            "vil_5_stack": 41,
            "pot": 42,
            "amount_to_call": 43,
            "pot_odds": 44,
            "previous_amount": 45,
            "previous_position": 46,
            "previous_action": 47,
            "previous_bet_is_blind": 48,
            "next_player": 49,
            "current_player": 50,
        }
        player_state_shape = 51

        global_state_mapping = {
            "player_1_hand_range": [0, 8],
            "player_2_hand_range": [8, 16],
            "player_3_hand_range": [16, 24],
            "player_4_hand_range": [24, 32],
            "player_5_hand_range": [32, 40],
            "player_6_hand_range": [40, 48],
            "board_range": [48, 58],
            "street": 58,
            "num_players": 59,
            "player_1_position": 60,
            "player_2_position": 61,
            "player_3_position": 62,
            "player_4_position": 63,
            "player_5_position": 64,
            "player_6_position": 65,
            "player_1_stack": 66,
            "player_2_stack": 67,
            "player_3_stack": 68,
            "player_4_stack": 69,
            "player_5_stack": 70,
            "player_6_stack": 71,
            "player_1_active": 72,
            "player_2_active": 73,
            "player_3_active": 74,
            "player_4_active": 75,
            "player_5_active": 76,
            "player_6_active": 77,
            "pot": 78,
            "amount_to_call": 79,
            "pot_odds": 80,
            "previous_amount": 81,
            "previous_position": 82,
            "previous_action": 83,
            "previous_bet_is_blind": 84,
            "last_agro_amount": 85,
            "last_agro_action": 86,
            "last_agro_position": 87,
            "last_agro_bet_is_blind": 88,
            "next_player": 89,
            "current_player": 90,
        }
        global_state_shape = 91
    elif game_type == GameTypes.HOLDEM:
        raise NotImplementedError
    elif game_type == GameTypes.OMAHA_HI_LO:
        raise NotImplementedError
    elif game_type == 'BIG_O':
        raise NotImplementedError
    else:
        raise NotImplementedError
    return player_state_mapping, player_state_shape, global_state_mapping, global_state_shape




from pokerrl_env.datatypes import BetLimits, GameTypes, Player, Positions
from pokerrl_env.utils import is_next_player_the_aggressor,return_deck
import pytest
from pokerrl_env.config import Config
from pokerrl_env.transition import init_state
import numpy as np


def test_default_values():
    config = Config()
    assert config.game_type == GameTypes.OMAHA_HI
    assert config.num_players == 2
    assert config.bet_limit == BetLimits.POT_LIMIT
    assert config.betsizes == (1, 0.9, 0.75, 0.67, 0.5, 0.33, 0.25, 0.1)
    assert config.blinds == (0.5,1)
    assert config.stack_sizes == 100

def test_custom_values():
    config = Config(game_type=GameTypes.OMAHA_HI, num_players=4, bet_limit=BetLimits.POT_LIMIT,
                    betsizes=(1, 0.75, 0.5, 0.25), blinds=(1, 0.5), stack_sizes=500)
    assert config.game_type == GameTypes.OMAHA_HI
    assert config.num_players == 4
    assert config.bet_limit == BetLimits.POT_LIMIT
    assert config.betsizes == (1, 0.75, 0.5, 0.25)
    assert config.blinds == (1, 0.5)
    assert config.stack_sizes == 500

def test_init_state():
    config = Config(num_players=4)
    state,done,winnings,action_mask = init_state(config)
    assert state.shape == (2,91)
    assert isinstance(state, np.ndarray)

def test_return_deck():
    deck = return_deck()
    assert len(deck) == 52
    assert isinstance(deck, list)

def test_is_next_player_the_aggressor():
    active_players = [Player(Positions.BIG_BLIND, 100, 1, [8, 14], 5, 10),Player(Positions.SMALL_BLIND, 100, 1, [8, 14], 5, 10),Player(Positions.DEALER, 100, 1, [8, 14], 5, 10),]
    current_player = 2
    last_agro_player = 1
    result = is_next_player_the_aggressor(active_players, current_player, last_agro_player)
    assert result == False


def test_is_next_player_the_aggressor2():
    active_players = [Player(Positions.BIG_BLIND, 100, 1, [8, 14], 5, 10),Player(Positions.SMALL_BLIND, 100, 1, [8, 14], 5, 10)]
    current_player = 2
    last_agro_player = 6
    result = is_next_player_the_aggressor(active_players, current_player, last_agro_player)
    assert result == True

def test_is_next_player_the_aggressor3():
    active_players = [Player(Positions.SMALL_BLIND, 100, 1, [8, 14], 5, 10),Player(Positions.BIG_BLIND, 100, 1, [8, 14], 5, 10)]
    current_player = 2
    last_agro_player = 6
    result = is_next_player_the_aggressor(active_players, current_player, last_agro_player)
    assert result == True

def test_is_next_player_the_aggressor4():
    active_players = [Player(Positions.SMALL_BLIND, 100, 1, [8, 14], 5, 10),Player(Positions.BIG_BLIND, 100, 1, [8, 14], 5, 10)]
    current_player = 2
    last_agro_player = 5
    result = is_next_player_the_aggressor(active_players, current_player, last_agro_player)
    assert result == True


def test_is_next_player_the_aggressor5():
    active_players = [Player(Positions.SMALL_BLIND, 100, 1, [8, 14], 5, 10),Player(Positions.DEALER, 100, 1, [8, 14], 5, 10)]
    current_player = 6
    last_agro_player = 2
    result = is_next_player_the_aggressor(active_players, current_player, last_agro_player)
    assert result == False


def test_is_next_player_the_aggressor6():
    active_players = [Player(Positions.UTG, 100, 1, [8, 14], 5, 10),Player(Positions.DEALER, 100, 1, [8, 14], 5, 10)]
    current_player = 6
    last_agro_player = 2
    result = is_next_player_the_aggressor(active_players, current_player, last_agro_player)
    assert result == True


def test_is_next_player_the_aggressor7():
    active_players = [Player(Positions.SMALL_BLIND, 100, 1, [8, 14], 5, 10),Player(Positions.DEALER, 100, 1, [8, 14], 5, 10)]
    current_player = 1
    last_agro_player = 2
    result = is_next_player_the_aggressor(active_players, current_player, last_agro_player)
    assert result == True



import pytest
import numpy as np
from pokerrl_env.transition import *
from pokerrl_env.config import Config
from pokerrl_env.datatypes import POSITION_TO_SEAT, SEAT_TO_POSITION, StateActions, Street
from pokerrl_env.utils import calculate_pot_limit_betsize, readable_card_to_int
from pokerrl_env.transition import get_action_mask,players_finished,step_state,init_state

config = Config(num_players=6)

@pytest.fixture
def river_state():
    global_state = np.zeros(config.global_state_shape, dtype=np.int32)
    for i in range(1, 4):
        global_state[config.global_state_mapping[f'player_{i}_stack']] = 100
        global_state[config.global_state_mapping[f'player_{i}_active']] = 1
        global_state[config.global_state_mapping[f'player_{i}_position']] = i

    global_state[config.global_state_mapping['last_agro_amount']] = 3
    global_state[config.global_state_mapping['last_agro_action']] = 3
    global_state[config.global_state_mapping['last_agro_position']] = 2
    global_state[config.global_state_mapping['last_agro_bet_is_blind']] = 0
    global_state[config.global_state_mapping['street']] = Street.RIVER
    global_state[config.global_state_mapping['current_player']] = 1
    global_state[config.global_state_mapping['current_player']] = 2
    global_state[config.global_state_mapping['pot']] = 50
    return global_state

def test_clear_last_agro_action():
    global_state = np.zeros(config.global_state_shape)
    global_state[config.global_state_mapping['last_agro_action']] = 1
    global_state[config.global_state_mapping['last_agro_position']] = 1
    global_state[config.global_state_mapping['last_agro_amount']] = 1
    global_state[config.global_state_mapping['last_agro_bet_is_blind']] = 1
    clear_last_agro_action(global_state, config)
    assert global_state[config.global_state_mapping['last_agro_action']] == 0
    assert global_state[config.global_state_mapping['last_agro_position']] == 0
    assert global_state[config.global_state_mapping['last_agro_amount']] == 0
    assert global_state[config.global_state_mapping['last_agro_bet_is_blind']] == 0

def test_players_finished():
    global_state = np.zeros(config.global_state_shape)
    global_state[config.global_state_mapping['player_1_active']] = 1
    global_state[config.global_state_mapping['player_1_stack']] = 1
    global_state[config.global_state_mapping['player_2_active']] = 1
    global_state[config.global_state_mapping['player_2_stack']] = 1
    assert not players_finished(global_state, config)
    global_state[config.global_state_mapping['player_2_active']] = 0
    assert players_finished(global_state, config)
    global_state[config.global_state_mapping['player_2_active']] = 1
    global_state[config.global_state_mapping['player_2_stack']] = 0
    assert players_finished(global_state, config)

def test_order_players_by_street():
    global_state = np.zeros(config.global_state_shape)
    global_state[config.global_state_mapping['street']] = Street.PREFLOP
    for i in range(1, 7):
        global_state[config.global_state_mapping[f'player_{i}_stack']] = 100
        global_state[config.global_state_mapping[f'player_{i}_active']] = 1
        global_state[config.global_state_mapping[f'player_{i}_position']] = i
    active_players = order_players_by_street(global_state, config)
    assert [p.position for p in active_players] == [3, 4, 5, 6, 1, 2]
    assert len(active_players) == 6
    global_state[config.global_state_mapping['street']] = Street.FLOP
    active_players = order_players_by_street(global_state, config)
    assert [p.position for p in active_players] == [1, 2, 3, 4, 5, 6]

def test_increment_players():
    global_state = np.zeros(config.global_state_shape)
    global_state[config.global_state_mapping['street']] = Street.FLOP
    current_player = 1
    for i in range(1, 7):
        global_state[config.global_state_mapping[f'player_{i}_stack']] = 100
        global_state[config.global_state_mapping[f'player_{i}_active']] = 1
        global_state[config.global_state_mapping[f'player_{i}_position']] = i
    global_state[config.global_state_mapping['current_player']] = current_player
    global_state[config.global_state_mapping['next_player']] = 2
    active_players = order_players_by_street(global_state, config)
    print('active_players', active_players)
    increment_players(global_state, active_players, current_player, config)
    assert global_state[config.global_state_mapping['current_player']] == 2
    assert global_state[config.global_state_mapping['next_player']] == 3

def test_new_street_player_order():
    global_state = np.zeros(config.global_state_shape)
    for i in range(1, 7):
        global_state[config.global_state_mapping[f'player_{i}_stack']] = 100
        global_state[config.global_state_mapping[f'player_{i}_active']] = 1
        global_state[config.global_state_mapping[f'player_{i}_position']] = i
    global_state[config.global_state_mapping['street']] = Street.FLOP
    new_street_player_order(global_state, config)
    assert global_state[config.global_state_mapping['current_player']] == 1
    assert global_state[config.global_state_mapping['next_player']] == 2

def test_init_state():
    config = Config(num_players=2)
    global_states,done,winnings,action_mask = init_state(config)
    assert global_states[-1,config.global_state_mapping['last_agro_amount']] == 1
    assert global_states[-1,config.global_state_mapping['player_6_stack']] == 99.5
    assert global_states[-1,config.global_state_mapping['player_2_stack']] == 99
    assert global_states[-1,config.global_state_mapping['last_agro_action']] == 4

def test_step_state():
    global_states,done,winnings,action_mask = init_state(config)
    global_states,done,winnings,action_mask = step_state(global_states, ModelActions.FOLD, config)
    assert global_states.shape == (3,config.global_state_shape)
    assert global_states[2,config.global_state_mapping['street']] == Street.PREFLOP
    assert global_states[2,config.global_state_mapping['previous_amount']] == 0
    assert global_states[2,config.global_state_mapping['previous_action']] == StateActions.FOLD
    assert global_states[2,config.global_state_mapping['previous_position']] == POSITION_TO_SEAT['UTG']

def test_classify_action():
    config = Config()
    action, betsize = classify_action(ModelActions.FOLD, 0, 100, 20,  StateActions.CALL+1, 60, config)
    assert action == FOLD and betsize == 0

    action, betsize = classify_action(ModelActions.CHECK, 0, 100, 20,  StateActions.CALL+1, 60, config)
    assert action == CHECK and betsize == 0

    action, betsize = classify_action(ModelActions.CALL, 0, 100, 20,  StateActions.CALL+1, 60, config)
    assert action == CALL and betsize == 20

    action, betsize = classify_action(ModelActions.CALL+1, 0, 100, 0, StateActions.CHECK, 60, config)
    assert action == BET and betsize == 60

    action, betsize = classify_action(ModelActions.CALL+1, 0, 100, 20, StateActions.CALL+1, 60, config)
    assert action == RAISE and betsize == 100


def test_classify_action_in_game():
    config = Config(num_players=2)
    global_state,done,winnings,action_mask = init_state(config)
    amount_invested = 0.5
    action, betsize = classify_action(ModelActions.FOLD, amount_invested, global_state[-1,config.global_state_mapping[f'player_1_stack']], global_state[-1,config.global_state_mapping['last_agro_amount']],global_state[-1,config.global_state_mapping['last_agro_action']],global_state[-1,config.global_state_mapping['pot']],config)
    assert action == FOLD and betsize == 0

    global_state,done,winnings,action_mask = init_state(config)
    amount_invested = 0.5
    action, betsize = classify_action(ModelActions.CALL, amount_invested, global_state[-1,config.global_state_mapping[f'player_1_stack']], global_state[-1,config.global_state_mapping['last_agro_amount']],global_state[-1,config.global_state_mapping['last_agro_action']],global_state[-1,config.global_state_mapping['pot']],config)
    assert action == CALL and betsize == 0.5

    action, betsize = classify_action(ModelActions.CALL, 0, 100, 20, 4, 60, config)
    assert action == CALL and betsize == 20

    # action, betsize = classify_action(4, 0, 100, 20, 0, 60, config)
    # assert action == BET and betsize == 60

    # action, betsize = classify_action(4, 0, 100, 20, 4, 60, config)
    # assert action == RAISE and betsize == 100


def test_blind_init():
    config = Config(num_players=2)
    global_state,done,winnings,action_mask = init_state(config)
    print(global_state[:,config.global_state_mapping['last_agro_amount']])
    assert global_state[0,config.global_state_mapping['last_agro_amount']] == 0.5
    assert global_state[-1,config.global_state_mapping['last_agro_amount']] == 1
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.DEALER
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.BIG_BLIND

def test_full_game_checked_to_river():
    config = Config(num_players=2)
    global_state,done,winnings,action_mask = init_state(config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.DEALER
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.DEALER
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.DEALER
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.BIG_BLIND
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.DEALER
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.,2.,3.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.DEALER
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.BIG_BLIND
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.,2.,3.,3.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.DEALER
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.,2.,3.,3.,3.,4.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.,2.,3.,3.,3.,4.,4.]))
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_total_amount_invested[2] == 1
    assert player_total_amount_invested[6] == 1
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.,2.,3.,3.,3.,4.,4.,4.]))
    assert global_state[-1,config.global_state_mapping['pot']] == 2
    if winnings[2]['hand_value'] < winnings[6]['hand_value']:
        assert winnings[2]['result'] == 1
        assert winnings[6]['result'] == -1
    elif winnings[2]['hand_value'] > winnings[6]['hand_value']:
        assert winnings[2]['result'] == -1
        assert winnings[6]['result'] == 1
    else:
        assert winnings[2]['result'] == 0
        assert winnings[6]['result'] == 0


def test_full_game_with_raise_to_river():
    config = Config(num_players=2)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    print(global_state[:,config.global_state_mapping['street']])
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.,2.,3.,3.,3.,4.,4.,4.]))
    assert global_state[-1,config.global_state_mapping['pot']] == 6
    assert done == True


def test_full_game_with_raise_to_allin():
    config = Config(num_players=2,stack_sizes=9)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config)
    print(global_state[:,config.global_state_mapping['street']])
    assert global_state[-1,config.global_state_mapping['pot']] == 18
    assert done == True


def test_full_game_with_raise_to_allin27():
    config = Config(num_players=2,stack_sizes=27)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config)
    print(global_state[:,config.global_state_mapping['street']])
    assert global_state[-1,config.global_state_mapping['pot']] == 54
    assert done == True


def test_full_game_2p_allin():
    config = Config(num_players=2,stack_sizes=100)
    comparison_mask = np.zeros(11)
    comparison_mask[0] = 1
    comparison_mask[2] = 1
    global_state,done,winnings,action_mask = init_state(config)
    assert global_state[-1,config.global_state_mapping['pot']] == 1.5
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # SB
    assert global_state[-1,config.global_state_mapping['pot']] == 4
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # BB
    assert global_state[-1,config.global_state_mapping['pot']] == 12
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # SB betsize of 9
    assert global_state[-1,config.global_state_mapping['pot']] == 18
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config) # BB
    assert global_state[-1,config.global_state_mapping['pot']] == 18
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config) # SB
    assert global_state[-1,config.global_state_mapping['pot']] == 18
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # BB betsize of 18, 73 left
    assert global_state[-1,config.global_state_mapping['pot']] == 36
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # SB call of 18, 73 left
    assert global_state[-1,config.global_state_mapping['pot']] == 54
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)# BB betsize of 54. 19 left
    assert global_state[-1,config.global_state_mapping['pot']] == 108
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # SB betsize of 81 allin
    assert global_state[-1,config.global_state_mapping['pot']] == 181
    assert np.array_equal(action_mask,comparison_mask)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # BB call of 19
    assert global_state[-1,config.global_state_mapping['pot']] == 200
    assert done == True

def test_full_game_with_3p_raise_to_allin27():
    config = Config(num_players=3,stack_sizes=27)
    comparison_mask = np.zeros(11)
    comparison_mask[0] = 1
    comparison_mask[2] = 1
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # 3.5
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # 11.5
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # 27
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.]))
    assert np.array_equal(action_mask,comparison_mask)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # 23.5
    assert np.array_equal(action_mask,comparison_mask)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # 15.5
    
    assert global_state[-1,config.global_state_mapping['player_1_stack']] == 0
    assert global_state[-1,config.global_state_mapping['player_2_stack']] == 0
    assert global_state[-1,config.global_state_mapping['player_6_stack']] == 0
    assert global_state[-1,config.global_state_mapping['pot']] == 81
    assert done == True


def test_full_game_with_3p_raise_to_allin100():
    config = Config(num_players=3,stack_sizes=100)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) #BTN 3.5
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) #SB 3.5
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # BB 14
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # BTN 10.5
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 14, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[1] == 3.5, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 14, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[1] == 3.5, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 14, f'{player_total_amount_invested}'

    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # SB 10.5
    # new street
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[1] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 0, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[1] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 14, f'{player_total_amount_invested}'
    assert global_state[-1,config.global_state_mapping['pot']] == 42
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config) # SB 0
    assert global_state[-1,config.global_state_mapping['pot']] == 42
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config) # BB 0
    assert global_state[-1,config.global_state_mapping['pot']] == 42
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[1] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 0, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[1] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 14, f'{player_total_amount_invested}'
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # BTN 42
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[1] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 42, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[1] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 56, f'{player_total_amount_invested}'
    assert global_state[-1,config.global_state_mapping['pot']] == 84
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.FOLD, config) # SB 0
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[1] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 42, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[1] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 56, f'{player_total_amount_invested}'
    assert global_state[-1,config.global_state_mapping['pot']] == 84
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # BB 86
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 86, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[1] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 42, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 100, f'{player_total_amount_invested}'
    assert player_total_amount_invested[1] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 56, f'{player_total_amount_invested}'
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # BTN 44
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 86, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[1] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 86, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 100, f'{player_total_amount_invested}'
    assert player_total_amount_invested[1] == 14, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 100, f'{player_total_amount_invested}'
    
    assert global_state[-1,config.global_state_mapping['player_1_stack']] == 86
    assert global_state[-1,config.global_state_mapping['player_2_stack']] == 0
    assert global_state[-1,config.global_state_mapping['player_6_stack']] == 0
    assert global_state[-1,config.global_state_mapping['pot']] == 214
    assert done == True



def test_full_game_with_raise_to_river_3_players():
    config = Config(num_players=3)
    global_state,done,winnings,action_mask = init_state(config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.DEALER
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.SMALL_BLIND
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.SMALL_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.BIG_BLIND
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.DEALER
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.]))
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.SMALL_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.BIG_BLIND
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.DEALER
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.DEALER
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.SMALL_BLIND
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.SMALL_BLIND
    assert global_state[-1,config.global_state_mapping['next_player']] == Positions.BIG_BLIND
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.,2.,3.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.,2.,3.,3.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.,2.,3.,3.,3.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.,2.,3.,3.,3.,3.,4.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.,2.,3.,3.,3.,3.,4.,4.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert done == False
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.,2.,3.,3.,3.,3.,4.,4.,4.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config)
    assert done == True
    print(global_state[:,config.global_state_mapping['street']])
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.,2.,2.,2.,3.,3.,3.,3.,4.,4.,4.,4.]))


def test_full_game_with_3p_BTN_Call():
    config = Config(num_players=3,stack_sizes=100)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # 1
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # 11.5
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config) # 27
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # SB 
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # BB call
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.DEALER


def test_full_game_with_3p_BTN_Call_SB_FOLD():
    config = Config(num_players=3,stack_sizes=100)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # 1
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.FOLD, config) # 11.5
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CHECK, config) # 27
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,1.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # SB 
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # BB call
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND


def test_full_game_with_2p_BTN_Call_SB_FOLD():
    config = Config(num_players=2,stack_sizes=100)
    mask = np.ones((config.num_actions))
    mask[1] = 0
    mask[3] = 0
    mask[4] = 0
    global_state,done,winnings,action_mask = init_state(config)
    assert np.array_equal(global_state[:,config.global_state_mapping['current_player']],np.array([2.,6.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL+1, config) # SB 3
    assert np.array_equal(global_state[:,config.global_state_mapping['current_player']],np.array([2.,6.,2.]))
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config) # BB 3
    assert np.array_equal(global_state[:,config.global_state_mapping['current_player']],np.array([2.,6.,2.,2.,2.]))
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.]))
    print('current_player',global_state[:,config.global_state_mapping['current_player']])
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 0, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 0, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 3, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 3, f'{player_total_amount_invested}'
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL+1, config) # BB 6
    assert np.array_equal(global_state[:,config.global_state_mapping['current_player']],np.array([2.,6.,2.,2.,2.,6.]))
    assert np.array_equal(global_state[:,config.global_state_mapping['street']],np.array([1.,1.,1.,1.,2.,2.]))
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 6, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 0, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 9, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 3, f'{player_total_amount_invested}'
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # SB 24
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 6, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 24, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 9, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 27, f'{player_total_amount_invested}'
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # BB 78
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 78, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 24, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 81, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 27, f'{player_total_amount_invested}'
    assert np.array_equal(action_mask,mask)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 3, config) # SB 100

    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 78, f'{player_amount_invested_per_street}'
    assert player_amount_invested_per_street[6] == 97, f'{player_amount_invested_per_street}'
    assert player_total_amount_invested[2] == 81, f'{player_total_amount_invested}'
    assert player_total_amount_invested[6] == 100, f'{player_total_amount_invested}'
    assert global_state[-1,config.global_state_mapping['current_player']] == Positions.BIG_BLIND

### Action Mask


def test_fold_allowed_vs_bet_raise(river_state):
    player_total_amount_invested = {1: 50, 2: 50, 3: 50, 4: 50, 5: 50, 6: 50}
    action_mask = get_action_mask(river_state, player_total_amount_invested, config)
    print(action_mask)
    assert action_mask[0] == 1, "Fold should be allowed vs bet/raise"


def test_check_allowed(river_state):
    player_total_amount_invested = {1: 50, 2: 50, 3: 50, 4: 50, 5: 50, 6: 50}
    river_state[config.global_state_mapping['last_agro_amount']] = 0
    river_state[config.global_state_mapping['last_agro_action']] = 0
    river_state[config.global_state_mapping['last_agro_position']] = 0
    action_mask = get_action_mask(river_state, player_total_amount_invested, config)
    assert action_mask[1] == 1, "Check should be allowed when the action is unopened"


def test_call_allowed(river_state):
    player_total_amount_invested = {1: 50, 2: 50, 3: 50, 4: 50, 5: 50, 6: 50}
    action_mask = get_action_mask(river_state, player_total_amount_invested, config)
    assert action_mask[2] == 1, "Call should be allowed when the last aggressive action is a bet or raise"


def test_bet_sizes_allowed():
    config = Config(num_players=2,stack_sizes=27)
    global_state,done,winnings,action_mask = init_state(config)

    # player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    # action_mask = get_action_mask(river_state, player_total_amount_invested, config)
    expected_action_mask = np.ones(config.num_actions, dtype=int)
    expected_action_mask[1] = 0 # no check
    assert np.array_equal(action_mask, expected_action_mask), "Bet sizes should be allowed based on stack and pot"

def test_preflop_fold():
    config = Config(num_players=2)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.FOLD, config)
    assert done == True
    assert winnings[6]['result'] == -0.5
    assert winnings[2]['result'] == 0.5
    assert winnings[6]['hand'] == []


def test_call_fold_only():
    config = Config(num_players=2,stack_sizes=9)
    global_state,done,winnings,action_mask = init_state(config)
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # 3
    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config) # 9
    assert np.array_equal(action_mask,np.array([1,0,1,0,0,0,0,0,0,0,0]))
    assert done == False

### TEST POT CALCS ###

def test_calculate_pot_limit_betsize_2_player_SB():
    config = Config(num_players=2,stack_sizes=9)
    last_agro_action = StateActions.CALL + 1
    last_agro_amount = 1
    action = ModelActions.CALL + 1
    pot = 1.5
    player_street_total = 0.5
    player_stack = 9
    action_str,betsize = calculate_pot_limit_betsize(last_agro_action, last_agro_amount, config, action, pot, player_street_total, player_stack)
    assert action_str == RAISE
    assert betsize == 3

def test_calculate_pot_limit_betsize_2_player_BB():
    config = Config(num_players=2,stack_sizes=9)
    last_agro_action = StateActions.CALL + 1
    last_agro_amount = 3
    action = ModelActions.CALL + 1
    pot = 4
    player_street_total = 1
    player_stack = 9
    action_str,betsize = calculate_pot_limit_betsize(last_agro_action, last_agro_amount, config, action, pot, player_street_total, player_stack)
    assert action_str == RAISE
    assert betsize == 9

def test_calculate_pot_limit_betsize_2_player_SB_3bet():
    config = Config(num_players=2,stack_sizes=27)
    last_agro_action = StateActions.CALL + 1
    last_agro_amount = 9
    action = ModelActions.CALL + 1
    pot = 12
    player_street_total = 3
    player_stack = 27
    action_str,betsize = calculate_pot_limit_betsize(last_agro_action, last_agro_amount, config, action, pot, player_street_total, player_stack)
    assert action_str == RAISE
    assert betsize == 27

def test_calculate_pot_limit_betsize_2_player_BB_4bet():
    config = Config(num_players=2,stack_sizes=27)
    last_agro_action = StateActions.CALL + 1
    last_agro_amount = 27
    action = ModelActions.CALL + 1
    pot = 36
    player_street_total = 9
    player_stack = 81
    action_str,betsize = calculate_pot_limit_betsize(last_agro_action, last_agro_amount, config, action, pot, player_street_total, player_stack)
    assert action_str == RAISE
    assert betsize == 81

def test_calculate_pot_limit_betsize_3_player():
    config = Config(num_players=3,stack_sizes=9)
    last_agro_action = StateActions.CALL + 1
    last_agro_amount = 1
    action = ModelActions.CALL + 1
    pot = 1.5
    player_street_total = 0
    player_stack = 9
    action_str,betsize = calculate_pot_limit_betsize(last_agro_action, last_agro_amount, config, action, pot, player_street_total, player_stack)
    assert action_str == RAISE
    assert betsize == 3.5

def test_calculate_pot_limit_betsize_3_player_SB_3bet():
    config = Config(num_players=3,stack_sizes=20)
    last_agro_action = StateActions.CALL + 1
    last_agro_amount = 3.5
    action = ModelActions.CALL + 1
    pot = 5
    player_street_total = 0.5
    player_stack = 20
    action_str,betsize = calculate_pot_limit_betsize(last_agro_action, last_agro_amount, config, action, pot, player_street_total, player_stack)
    assert action_str == RAISE
    assert betsize == 11.5

### TEST RETURN INVESTMENTS

def test_return_investments():
    config = Config(num_players=2,stack_sizes=27)
    global_state,done,winnings,action_mask = init_state(config)

    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 1
    assert player_amount_invested_per_street[6] == 0.5
    assert player_total_amount_invested[2] == 1
    assert player_total_amount_invested[6] == 0.5

    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)


    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 1
    assert player_amount_invested_per_street[6] == 3
    assert player_total_amount_invested[2] == 1
    assert player_total_amount_invested[6] == 3

    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)
    
    
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 9
    assert player_amount_invested_per_street[6] == 3
    assert player_total_amount_invested[2] == 9
    assert player_total_amount_invested[6] == 3

    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL + 1, config)


    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    assert player_amount_invested_per_street[2] == 9
    assert player_amount_invested_per_street[6] == 27
    assert player_total_amount_invested[2] == 9
    assert player_total_amount_invested[6] == 27

    global_state,done,winnings,action_mask = step_state(global_state, ModelActions.CALL, config)
    assert done == True
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_state, config)
    # assert player_amount_invested_per_street[2] == 27
    # assert player_amount_invested_per_street[6] == 27
    # assert player_total_amount_invested[2] == 27
    # assert player_total_amount_invested[6] == 27


import numpy as np
from pokerrl_env.datatypes import Positions
import pytest
from pokerrl_env.view import player_view, human_readable_view,return_board_cards
from pokerrl_env.config import Config
from pokerrl_env.transition import init_state


@pytest.fixture
def config():
    return Config(num_players=6)


@pytest.fixture
def initial_states(config):
    return init_state(config)


@pytest.fixture
def player_index():
    return 1

@pytest.fixture
def dealer_position():
    return Positions.DEALER

@pytest.fixture
def utg_position():
    return Positions.UTG

def test_return_board_cards(initial_states,config:Config):
    global_state,_,_,_ = initial_states
    board_cards = return_board_cards(global_state[-1],config)
    print(board_cards)
    assert np.array_equal(board_cards,np.zeros(10))
    state = global_state[-1]
    state[config.global_state_mapping['street']] = 2
    board_cards = return_board_cards(state,config)
    assert np.array_equal(board_cards[6:],np.zeros(4))
    state[config.global_state_mapping['street']] = 3
    board_cards = return_board_cards(state,config)
    assert np.array_equal(board_cards[8:],np.zeros(2))
    state[config.global_state_mapping['street']] = 4
    board_cards = return_board_cards(state,config)
    # the next line asserts board_cards has positive numbers only
    assert np.all(board_cards)

def test_player_view(initial_states,utg_position,config:Config):
    print('test_player_view',utg_position)
    global_state,_,_,_ = initial_states
    player_states = player_view(global_state, utg_position,config)
    print(player_states[:,:30])
    assert player_states.shape == (2, config.player_state_shape), "Player view should have a shape of (2,config.player_state_shape)."
    assert np.all(player_states[:, 20] == utg_position), f"Player index should be consistent in the player view. {player_states[:, 20]}, {utg_position}"


def test_human_readable_view(initial_states,config):
    dealer_position = Positions.DEALER
    global_state,_,_,_ = initial_states
    human_readable_states = human_readable_view(global_state, dealer_position,config)
    for state in human_readable_states:
        assert "hand_range" in state, "Hand range should be present in the human-readable view."
        assert "board_range" in state, "Board range should be present in the human-readable view."
        assert "street" in state, "Street should be present in the human-readable view."
        assert "num_players" in state, "Number of players should be present in the human-readable view."
        assert "hero_position" in state, "Hero position should be present in the human-readable view."




from .play import play_game
from .config import Config
from .view import player_view, human_readable_view, json_view
from .transition import step_state, init_state
from .datatypes import GameTypes,BetLimits,Positions
from .utils import return_current_player
from .game import Game
from .cardlib import encode, hand_rank



from pokerrl_env.transition import init_state, step_state
from pokerrl_env.config import Config
from pokerrl_env.utils import return_current_player
from pokerrl_env.view import player_view,json_view

class Game:
    def __init__(self,config:Config):
        self.config = config
        self.global_state = None
        self.done = None
        self.winnings = None
        self.action_mask = None

    def reset(self):
        """ Returns (state, reward, done, info) from the current player's perspective """
        self.global_state, self.done, self.winnings, self.action_mask = init_state(self.config)
        current_player = return_current_player(self.global_state,self.config)
        if self.config.is_server:
            return {"game_states":json_view(self.global_state,current_player,self.config), "done":self.done, "winnings":self.winnings, "action_mask":self.action_mask.tolist()}
        return self.global_state, self.done, self.winnings, self.action_mask

    def step(self,action):
        """ Returns (state, reward, done, info) from the current player's perspective """
        self.global_state, self.done, self.winnings, self.action_mask = step_state(self.global_state, action, self.config)
        current_player = return_current_player(self.global_state,self.config)
        if self.config.is_server:
            return {"game_states":json_view(self.global_state,current_player,self.config), "done":self.done, "winnings":self.winnings, "action_mask":self.action_mask.tolist()}
        return self.global_state, self.done, self.winnings, self.action_mask


import ctypes
from os import path
from sys import platform, argv
import pathlib

parent_dir = pathlib.Path(__file__).parent.parent.absolute()
release_dir = parent_dir / 'rusteval' / 'target' / 'release' 
if platform == "linux" or platform == "linux2":
    # linux
    lib = ctypes.cdll.LoadLibrary(path.join(release_dir, "librusteval.so"))
elif platform == "darwin":
    # OS X
    lib_path = path.join(release_dir, "librusteval.dylib")
    lib = ctypes.cdll.LoadLibrary(path.join(release_dir, "librusteval.dylib"))
else:
    # Windows...
    lib_path = path.join(release_dir, "librusteval.dll")
    lib = ctypes.cdll.LoadLibrary(path.join(release_dir, "librusteval.dll"))


# takes [2,'s'] or [2,1] and returns the cactus kev encoding for that card
def encode(card):
    """
    Ranks 1-13
    Suits 1-5
    """
    suitnum = card[1]
    if isinstance(card[1],str):
        suitnum = {'s': 1, 'h': 2,'d': 3,'c': 4}.get(card[1], 0)
    return lib.encode(ctypes.c_byte(card[0] - 1), ctypes.c_byte(suitnum - 1))

# takes a cactus kev encoded card and returns a list like [2,'s']
def decode(encoded):
    rank = ((encoded >> 8) & 0xF) + 2
    suit = ['','s','h','','d','','','','c'][(encoded >> 12) & 0xF]
    return [rank, suit]

# takes two 4-card hands and a 5-card board and
# returns 1 for hand1 wins, -1 for hand2 wins, or 0 for tie
def winner(hand1, hand2, board):
    return lib.winner(long_array(hand1), long_array(hand2), long_array(board))

# takes two 2-card hands and a 5-card board and
# returns 1 for hand1 wins, -1 for hand2 wins, or 0 for tie
def holdem_winner(hand1, hand2, board):
    return lib.holdem_winner(long_array(hand1), long_array(hand2), long_array(board))

# takes a 4 card hand and a 5 card board and returns the best rank of the 60 possible combinations
def hand_rank(hand, board):
    return lib.hand_with_board_rank(long_array(hand), long_array(board))

def holdem_hand_rank(hand, board):
    return lib.holdem_hand_with_board_rank(long_array(hand), long_array(board))
# for converting an array to a c array for passing to rust
def long_array(arr):
    return (ctypes.c_long * len(arr))(*arr)

def rank(hand):
    return lib.rank(*hand)

# example usage:
"""
SUITS: Accepts strings [s,c,h,d], or numbers 1-4.
RANKS: Accepts numbers 2-14 inclusive.
>>> hand = [[7, 's'], [5, 'c'], [14, 'h'], [10, 'h']]
>>> hand_en = [encode(c) for c in hand]
>>> hand2 = [[14, 'c'], [2, 's'], [2, 'd'], [11, 's']]
>>> hand2_en = [encode(c) for c in hand]
>>> board = [[10, 'c'], [2, 'h'], [4, 'c'], [13, 'c'], [4, 'h']]
>>> en_board = [encode(c) for c in board]
>>> winner(hand_en,hand2_en,en_board)
>>> rank(hand_en)
>>> hand_rank(hand_en,en_board)
>>> holdem_hand = [[14, 'c'], [2, 's']]
>>> holdem_hand_en = [encode(c) for c in holdem_hand]
>>> holdem_hand_rank(holdem_hand_en,en_board)
"""



RAISE = "Raises"
BET = "Bets"
FOLD = "Folds"
CALL = "Calls"
CHECK = "Checks"
BLIND = "blind"

class GameTypes:
    HOLDEM = 'Holdem'
    OMAHA_HI = 'OmahaHi'
    OMAHA_HI_LO = 'OmahaHiLo'

class Positions:
    SMALL_BLIND = 1
    BIG_BLIND = 2
    UTG = 3
    UTG_1 = 4
    UTG_2 = 5
    DEALER = 6

class Player:
    def __init__(self,position,stack,active,hand=None,hand_value=None,total_invested=0):
        self.position = int(position)
        self.stack = stack
        self.active = active
        self.hand = hand
        self.hand_value = hand_value
        self.total_invested = total_invested

    def __repr__(self):
        return f'Position:{self.position}, Stack:{self.stack}, Active:{self.active}, Hand:{self.hand}'

class BetLimits:
    POT_LIMIT = 'Pot limit'
    NO_LIMIT = 'No limit'
    FIXED_LIMIT = 'Fixed limit'

class StateActions:
    PADDED = 0
    FOLD = 1
    CHECK = 2
    CALL = 3

class ModelActions:
    FOLD = 0
    CHECK = 1
    CALL = 2


class Street:
    PADDED = 0
    PREFLOP = 1
    FLOP = 2
    TURN = 3
    RIVER = 4

INT_TO_STREET = {
    0: '_',
    1: 'PREFLOP',
    2: 'FLOP',
    3: 'TURN',
    4: 'RIVER'
}
BOARD_CARDS_GIVEN_STREET = {
    0: 0,
    1: 0,
    2: 6,
    3: 8,
    4: 10,
}

POSITION_TO_SEAT = {
    None: 0,
    'Small Blind': 1,
    'Big Blind': 2,
    'UTG':3,
    'UTG+1':4,
    'UTG+2':5,
    'Dealer':6
}
SEAT_TO_POSITION = {v: k for k, v in POSITION_TO_SEAT.items()}

rake_cap = {
    "low": {
        2: 0.5,
        3: 1,
        4: 1,
        5: 2,
        6: 2,
        7: 2,
        8: 2,
        9: 2,
    },
    "high": {
        2: 1,
        3: 2,
        4: 3,
        5: 3,
        6: 4,
        7: 4,
        8: 4,
        9: 4,
    },
}


ranks = ["2", "3", "4", "5", "6", "7", "8", "9", "T", "J", "Q", "K", "A"]
suits = ["c", "d", "h", "s"]
# card_to_int = {}
rank_to_int = {rank: i for i, rank in enumerate(ranks, start=1)}
rank_to_int["_"] = 0
suit_to_int = {suit: i for i, suit in enumerate(suits, start=1)}
suit_to_int["_"] = 0
# idx = 1  # zero padded
# for rank in range(0, 13):
#     for suit in range(0, 4):
#         card_to_int[ranks[rank] + suits[suit]] = idx
#         idx += 1
# int_to_card = {v: k for k, v in card_to_int.items()}
int_to_rank = {v: k for k, v in rank_to_int.items()}
int_to_suit = {v: k for k, v in suit_to_int.items()}
int_to_suit[0] = "-"
int_to_rank[0] = "-"
preflop_positions = ["UTG", "UTG+1", "UTG+2", "Dealer", "Small Blind", "Big Blind"]
preflop_order = {POSITION_TO_SEAT[p]: i for i,p in enumerate(preflop_positions,start=1)} # values change preflop to postflop
postflop_positions = [
    "Small Blind",
    "Big Blind",
    "UTG",
    "UTG+1",
    "UTG+2",
    "Dealer",
]
postflop_order = {POSITION_TO_SEAT[p]: POSITION_TO_SEAT[p] for p in postflop_positions}
PREFLOP_ORDER_BY_NUM_PLAYERS = {
    2: ['Dealer', 'Big Blind'],
    3: ['Dealer', 'Small Blind', 'Big Blind'],
    4: ['UTG','Dealer', 'Small Blind', 'Big Blind'],
    5: ['UTG','UTG+1','Dealer', 'Small Blind', 'Big Blind'],
    6: ['UTG','UTG+1','UTG+2','Dealer', 'Small Blind', 'Big Blind'],
}
POSTFLOP_ORDER_BY_NUM_PLAYERS = {
    2: ['Big Blind', 'Dealer'],
    3: ['Small Blind', 'Big Blind', 'Dealer'],
    4: ['Small Blind', 'Big Blind', 'UTG', 'Dealer'],
    5: ['Small Blind', 'Big Blind', 'UTG', 'UTG+1', 'Dealer'],
    6: ['Small Blind', 'Big Blind', 'UTG', 'UTG+1', 'UTG+2', 'Dealer'],
}
PLAYER_ORDER_BY_STREET = {
    Street.PREFLOP: preflop_order,
    Street.FLOP: postflop_order,
    Street.TURN: postflop_order,
    Street.RIVER: postflop_order,
}
PLAYERS_POSITIONS_DICT = {
    2: ["Dealer", "Big Blind"],
    3: ["Small Blind", "Big Blind", "Dealer"],
    4: ["Small Blind", "Big Blind", "UTG", "Dealer"],
    5: ["Small Blind", "Big Blind", "UTG", "UTG+1", "Dealer"],
    6: ["Small Blind", "Big Blind", "UTG", "UTG+1", "UTG+2", "Dealer"],
}
 
INT_POSITIONS_BY_NUM_PLAYERS = {
    2: [6, 2],
    3: [1, 2, 6],
    4: [1, 2, 5, 6],
    5: [1, 2, 4, 5, 6],
    6: [1, 2, 3, 4, 5, 6],
}


from typing import List
from pokerrl_env.datatypes import BET, RAISE, ModelActions, Player, StateActions, Street,Positions, int_to_rank, int_to_suit,rank_to_int,suit_to_int
from random import shuffle
import numpy as np

def return_deck():
    deck = []
    for suit in range(1,5):
        for rank in range(1,14):
            deck.append((rank,suit))
    shuffle(deck)
    return deck

def human_readable_cards(cards):
    readable_cards = []
    for i in range(0,len(cards),2):
        readable_cards.append([int_to_rank[cards[i]],int_to_suit[cards[i+1]]])
    return readable_cards

def readable_card_to_int(card):
    return [rank_to_int[card[0]],suit_to_int[card[1]]]

def return_current_player(global_states,config):
    return int(global_states[-1,config.global_state_mapping['current_player']])

def is_next_player_the_aggressor(active_players:List[Player], current_player:int, last_agro_position:int):
    """ for when the aggressor is allin and is not in active players. """
    current_player_index = [player.position for player in active_players].index(current_player)
    next_player_position = active_players[(current_player_index + 1) % len(active_players)].position
    if next_player_position < current_player:
        next_player_position + 6
    distance_to_agro = (last_agro_position - current_player) % 6
    distance_to_next = (next_player_position - current_player) % 6
    return distance_to_next > distance_to_agro
    

#### Action Mask Functions ####

def calculate_pot_limit_mask(global_state,config,pot,current_player_investment,current_player_stack):
    action_mask = np.zeros(config.num_actions, dtype=int)  # +2 for check and fold
    if global_state[config.global_state_mapping["last_agro_action"]] > ModelActions.CALL:
        max_raise = (pot - current_player_investment) + (2 * global_state[config.global_state_mapping["last_agro_amount"]])
        max_bet = min(current_player_stack+current_player_investment,max_raise)
        if max_bet <= global_state[config.global_state_mapping["last_agro_amount"]]:
            max_bet = 0

        # check for special case preflop blind situation.
        if global_state[config.global_state_mapping["street"]] == Street.PREFLOP and \
            global_state[config.global_state_mapping["current_player"]] == Positions.BIG_BLIND and \
            global_state[config.global_state_mapping["last_agro_position"]] == Positions.BIG_BLIND:
            action_mask[1] = 1
        else:
            action_mask[0] = 1 # fold is possible
            action_mask[2] = 1 # call is possible
    else:
        # bet
        max_bet = min(current_player_stack,pot)
        action_mask[1] = 1 # check is possible
    if max_bet > 0:
        highest_bet_ratio = current_player_stack / max_bet
        for i,bet_ratio in enumerate(config.betsizes):
            if bet_ratio <= highest_bet_ratio:
                action_mask[i + 3:] = 1
                break
    return action_mask


def calculate_no_limit_mask(global_state,config,pot,current_player_investment,current_player_stack):
    action_mask = np.zeros(config.num_actions, dtype=int)  # +2 for check and fold
    max_bet = current_player_stack
    if global_state[config.global_state_mapping["last_agro_action"]] > ModelActions.CALL:
        # check for special case preflop blind situation.
        if global_state[config.global_state_mapping["street"]] == Street.PREFLOP and \
            global_state[config.global_state_mapping["current_player"]] == Positions.BIG_BLIND and \
            global_state[config.global_state_mapping["last_agro_position"]] == Positions.BIG_BLIND:
            action_mask[1] = 1
        else:
            action_mask[0] = 1 # fold is possible
            action_mask[2] = 1 # call is possible
    else:
        # bet
        action_mask[1] = 1 # check is possible
    if max_bet > 0:
        highest_bet_ratio = max_bet / pot
        for i,bet_ratio in enumerate(config.betsizes):
            if bet_ratio <= highest_bet_ratio:
                action_mask[i + 3:] = 1
                break
    return action_mask


def calculate_fixed_limit_mask(global_state,config,action,pot,current_player_investment,current_player_stack):
    """ TODO """
    action_mask = np.zeros(config.num_actions, dtype=int)  # +2 for check and fold
    if global_state[config.global_state_mapping["last_agro_action"]] > ModelActions.CALL:
        max_raise = (pot - current_player_investment) + (2 * global_state[config.global_state_mapping["last_agro_amount"]])
        max_bet = min(current_player_stack,max_raise)
        # check for special case preflop blind situation.
        if global_state[config.global_state_mapping["street"]] == Street.PREFLOP and \
            global_state[config.global_state_mapping["current_player"]] == Positions.BIG_BLIND and \
            global_state[config.global_state_mapping["last_agro_position"]] == Positions.BIG_BLIND:
            action_mask[1] = 1
        else:
            action_mask[0] = 1 # fold is possible
            action_mask[2] = 1 # call is possible
    else:
        # bet
        max_bet = min(current_player_stack,pot)
        action_mask[1] = 1 # check is possible
    if max_bet > 0:
        highest_bet_ratio = current_player_stack / max_bet
        for i,bet_ratio in enumerate(config.betsizes):
            if bet_ratio <= highest_bet_ratio:
                action_mask[i + 3:] = 1
                break
    return action_mask

#### Action Functions ####

def calculate_pot_limit_betsize(last_agro_action,last_agro_amount,config,action,pot,player_street_total,player_stack):
    if last_agro_action > StateActions.CALL:
        # find max raise -> call the raise, raise the pot
        max_raise = (2 * last_agro_amount) + (pot - player_street_total)
        min_raise = max(last_agro_amount + (pot - player_street_total),config.blinds[0])
        bet_ratio = config.betsizes[action - (ModelActions.CALL+1)]
        bet_amount = min(max(max_raise * bet_ratio,min_raise),player_stack + player_street_total)
        return RAISE, bet_amount
    else:
        # bet
        max_bet = pot
        bet_ratio = config.betsizes[action - (ModelActions.CALL+1)]
        bet_amount = min(max_bet * bet_ratio,player_stack)
        return BET, bet_amount
    
def calculate_no_limit_betsize(last_agro_action,last_agro_amount,config,action,pot,player_street_total,player_stack):
    if last_agro_action > StateActions.CALL:
        # find max raise -> call the raise, raise the pot
        min_raise = last_agro_amount + (pot - player_street_total)
        bet_ratio = config.betsizes[action - (ModelActions.CALL+1)]
        bet_amount = min(max(min_raise * bet_ratio,min_raise),player_stack + player_street_total)
        return RAISE, bet_amount
    else:
        # bet
        bet_ratio = config.betsizes[action - (ModelActions.CALL+1)]
        bet_amount = min(pot * bet_ratio,player_stack)
        return BET, bet_amount
    

def calculate_fixed_limit_betsize(last_agro_action,last_agro_amount,config,action,pot,player_street_total,player_stack):
    """ TODO """
    if last_agro_action > StateActions.CALL:
        # find max raise -> call the raise, raise the pot
        max_raise = (2 * last_agro_amount) + (pot - player_street_total)
        bet_ratio = config.betsizes[action - (ModelActions.CALL+1)]
        bet_amount = min(max_raise * bet_ratio,player_stack + player_street_total)
        return RAISE, bet_amount
    else:
        # bet
        max_bet = pot
        bet_ratio = config.betsizes[action - (ModelActions.CALL+1)]
        bet_amount = min(max_bet * bet_ratio,player_stack)
        return BET, bet_amount


import numpy as np
from pokerrl_env.config import Config
from pokerrl_env.datatypes import BOARD_CARDS_GIVEN_STREET, SEAT_TO_POSITION, INT_TO_STREET
from pokerrl_env.utils import human_readable_cards


def return_board_cards(global_state, config):
    num_board_cards = BOARD_CARDS_GIVEN_STREET[ global_state[config.global_state_mapping["street"]] ]
    if num_board_cards == 0:
        padded_board_cards = np.zeros(10)
    else:
        board_slice = global_state[ config.global_state_mapping["board_range"][ 0 ] : config.global_state_mapping["board_range"][1] ][:num_board_cards]
        padded_array = np.zeros(10 - num_board_cards)
        padded_board_cards = np.concatenate((board_slice,padded_array))
    return padded_board_cards


def convert_global_state_to_player_view(global_state, player_index, config):
    pmap = config.player_state_mapping
    gmap = config.global_state_mapping
    player_state = np.zeros(config.player_state_shape)
    vil_index = lambda x: x
    for i,position in enumerate(config.player_positions,start=1):

        # match player position
        if player_index == position:
            player_state[pmap["hand_range"][0] : pmap["hand_range"][1]] = global_state[ gmap[f"player_{position}_hand_range"][0] : gmap[f"player_{position}_hand_range"][1] ]
            player_state[pmap["hero_active"]] = global_state[gmap[f"player_{position}_active"]]
            player_state[pmap["hero_stack"]] = global_state[gmap[f"player_{position}_stack"]]
            player_state[pmap["hero_position"]] = global_state[gmap[f"player_{position}_position"]]
            vil_index = lambda x: x - 1
        else:
            player_state[pmap[f"vil_{vil_index(i)}_active"]] = global_state[ gmap[f"player_{position}_active"] ]
            player_state[pmap[f"vil_{vil_index(i)}_stack"]] = global_state[ gmap[f"player_{position}_stack"] ]
            player_state[pmap[f"vil_{vil_index(i)}_position"]] = global_state[ gmap[f"player_{position}_position"] ]
    # Copy over the board cards
    player_state[pmap["board_range"][0] : pmap["board_range"][1]] = return_board_cards( global_state,config )
    global_state[gmap["board_range"][0]]: global_state[gmap["board_range"][1]]
    # copy general state
    for key in [
        "pot",
        "amount_to_call",
        "pot_odds",
        "street",
        "num_players",
        "current_player",
        "previous_amount",
        "previous_position",
        "previous_action",
        "previous_bet_is_blind",
        "last_agro_amount",
        "last_agro_position",
        "last_agro_action",
        "last_agro_bet_is_blind",
        "next_player",
    ]:
        player_state[pmap[key]] = global_state[gmap[key]]
    return player_state


def player_view(global_states, player_index, config):
    assert player_index in list(range(1,7)), "Player index must be between 1 and 6"
    # assert global_states[-1, config.global_state_mapping["current_player"]] == player_index, "Player index does not match current player"
    player_states = [
        convert_global_state_to_player_view(global_states[i,:], player_index, config)
        for i in range(global_states.shape[0])
    ]
    return np.stack(player_states)

def flatten(data:list):
    return [item for sublist in data for item in sublist]

def display_readable_states(readable_states):
    print("Hand : ","".join(flatten(readable_states[-1]['hand_range'])))
    for state in readable_states:
        print(state)
        print("Board: ","".join(flatten(state['board_range'])))
        print("Pot: ",state['pot'])
        print("hero_stack: ",state['hero_stack'])
        print("vil_1_stack: ",state['vil_1_stack'])
        print("vil_2_stack: ",state['vil_2_stack'])
        print("last_agro_position: ",state['last_agro_position'])
        print("last_agro_amount: ",state['last_agro_amount'])
        print("last_agro_action: ",state['last_agro_action'])
        print("previous_amount: ",state['previous_amount'])
        print("previous_position: ",state['previous_position'])
        print("previous_action: ",state['previous_action'])

def human_readable_view(global_states, player_index, config,display=False):
    player_states = player_view(global_states, player_index, config)
    readable_states = []
    for player_state in player_states:
        human_readable = {}
        for key, value in config.player_state_mapping.items():
            if "hand_range" in key:
                cards = human_readable_cards(player_state[value[0] : value[1]])
                human_readable[key] = cards
            elif "_position" in key:
                human_readable[key] = SEAT_TO_POSITION[int(player_state[value])]
            elif "_action" in key:
                human_readable[key] = config.action_to_str[int(player_state[value])]
            elif "street" in key:
                human_readable[key] = INT_TO_STREET[int(player_state[value])]
            elif (
                "_stack" in key
                or "_active" in key in key
                or "num_players" in key
                or "pot" in key
                or "amount_to_call" in key
                or "pot_odds" in key
                or "previous_amount" in key
                or "previous_bet_is_blind" in key
                or "previous_amount" in key
                or "previous_bet_is_blind" in key
                or "next_player" in key
                or "last_agro_amount" in key
                or "last_agro_position" in key
                or "last_agro_action" in key
                or "last_agro_bet_is_blind" in key
            ):
                human_readable[key] = player_state[value]
            elif "board_range" in key:
                cards = human_readable_cards(player_state[value[0] : value[1]])
                human_readable[key] = cards
        readable_states.append(human_readable)
    if display:
        display_readable_states(readable_states)
    return readable_states

def json_view(global_states, player_index, config:Config):
    player_states = player_view(global_states, player_index, config)
    json_states = []
    for state in player_states:
        state_object = {
            'hero_cards'                :state[config.player_state_mapping['hand_range'][0]:config.player_state_mapping['hand_range'][1]].tolist(),
            'board_cards'               :state[config.player_state_mapping['board_range'][0]:config.player_state_mapping['board_range'][1]].tolist(),
            'street'                    :state[config.player_state_mapping['street']],
            'num_players'               :state[config.player_state_mapping['num_players']],
            'hero_position'             :state[config.player_state_mapping['hero_position']],
            'hero_active'               :state[config.player_state_mapping['hero_active']],
            'vil1_active'               :state[config.player_state_mapping['vil_1_active']],
            'vil2_active'               :state[config.player_state_mapping['vil_2_active']],
            'vil3_active'               :state[config.player_state_mapping['vil_3_active']],
            'vil4_active'               :state[config.player_state_mapping['vil_4_active']],
            'vil5_active'               :state[config.player_state_mapping['vil_5_active']],
            'vil1_position'             :state[config.player_state_mapping['vil_1_position']],
            'vil2_position'             :state[config.player_state_mapping['vil_2_position']],
            'vil3_position'             :state[config.player_state_mapping['vil_3_position']],
            'vil4_position'             :state[config.player_state_mapping['vil_4_position']],
            'vil5_position'             :state[config.player_state_mapping['vil_5_position']],
            'last_agro_amount'          :state[config.player_state_mapping['last_agro_amount']],
            'last_agro_action'          :state[config.player_state_mapping['last_agro_action']],
            'last_agro_position'        :state[config.player_state_mapping['last_agro_position']],
            'last_agro_is_blind'        :state[config.player_state_mapping['last_agro_bet_is_blind']],
            'hero_stack'                :state[config.player_state_mapping['hero_stack']],
            'vil1_stack'                :state[config.player_state_mapping['vil_1_stack']],
            'vil2_stack'                :state[config.player_state_mapping['vil_2_stack']],
            'vil3_stack'                :state[config.player_state_mapping['vil_3_stack']],
            'vil4_stack'                :state[config.player_state_mapping['vil_4_stack']],
            'vil5_stack'                :state[config.player_state_mapping['vil_5_stack']],
            'pot'                       :state[config.player_state_mapping['pot']],
            'amount_to_call'            :state[config.player_state_mapping['amount_to_call']],
            'pot_odds'                  :state[config.player_state_mapping['pot_odds']],
            'previous_amount'           :state[config.player_state_mapping['previous_amount']],
            'previous_position'         :state[config.player_state_mapping['previous_position']],
            'previous_action'           :state[config.player_state_mapping['previous_action']],
            'previous_bet_is_blind'     :state[config.player_state_mapping['previous_bet_is_blind']],
            "current_player"            :state[config.player_state_mapping['current_player']],
            "next_player"               :state[config.player_state_mapping['next_player']],
        }
        json_states.append(state_object)
    return json_states


from pokerrl_env.utils import readable_card_to_int
import pytest
import numpy as np
from typing import List, Tuple, Dict
from pokerrl_env.config import Config
from pokerrl_env.datatypes import Player, Street
from pokerrl_env.transition import get_pots, calculate_winnings, game_over

@pytest.fixture
def example_players():
    return [
        Player(1, 100, 1, [8, 14], 5, 100),
        Player(2, 0, 1, [4, 5], 3, 50),
        Player(3, 150, 1, [12, 9], 5, 100)
    ]

@pytest.fixture
def config2():
    return Config(num_players=2)

@pytest.fixture
def config3():
    return Config(num_players=3)

@pytest.fixture
def example_total_amount_invested():
    return {1: 100, 2: 50, 3: 100}

@pytest.fixture
def total_amount_invested_2():
    return {1: 100, 2: 50, 3: 100}


@pytest.fixture
def example_pots(example_players):
    return [
        [150, [example_players[0], example_players[1],example_players[2]]],
        [100, [example_players[0], example_players[2]]]
    ]

def test_simple_pot(example_players):
    pots = get_pots([example_players[0],example_players[2]], {1: 100, 3: 100})
    assert pots == [
        [200, [example_players[0],example_players[2]]]
    ]

def test_get_pots(example_players, example_total_amount_invested,example_pots):
    pots = get_pots(example_players, example_total_amount_invested)
    print('final pots', pots)
    assert pots == example_pots

def test_calculate_winnings(example_pots, example_players):
    winnings = calculate_winnings(example_pots, example_players)
    print('winnings', winnings)
    assert winnings == {
        1: {'hand': [8, 14], 'hand_value': 5, 'result': -50},
        2: {'hand': [4, 5], 'hand_value': 3, 'result': 100},
        3: {'hand': [12, 9], 'hand_value': 5, 'result': -50}
    }

# Add more tests for game_over function if needed

def test_game_over_no_side_pots(config2):
    global_state = np.zeros(config2.global_state_shape)
    # set stacks for two players
    board_cards = [('A', 's'), ('K', 's'), ('Q', 's'), ('J', 's'), ('T', 's')]
    board = [readable_card_to_int(card) for card in board_cards]
    hand1 = [('A', 'h'), ('K', 'h'), ('Q', 'h'), ('J', 'h')]
    hand1 = [readable_card_to_int(card) for card in hand1]
    hand2 = [('A', 'c'), ('K', 'c'), ('9', 's'), ('8', 's')]
    hand2 = [readable_card_to_int(card) for card in hand2]
    # flatten hand1 and hand2
    hand1 = [item for sublist in hand1 for item in sublist]
    hand2 = [item for sublist in hand2 for item in sublist]
    board = [item for sublist in board for item in sublist]
    hands = [hand1,hand2]
    print(board)
    print('hands', hands)
    print(global_state.shape)
    total_amount_invested = {6: 25, 2: 25}
    for i,position in enumerate([2,6]):
        global_state[config2.global_state_mapping[f'player_{position}_stack']] = 100
        global_state[config2.global_state_mapping[f'player_{position}_active']] = 1
        global_state[config2.global_state_mapping[f'player_{position}_position']] = position
        global_state[config2.global_state_mapping[f'player_{position}_hand_range'][0]:config2.global_state_mapping[f'player_{position}_hand_range'][1]] = hands[i]

    global_state[config2.global_state_mapping['board_range'][0]:config2.global_state_mapping['board_range'][1]] = board
    global_state[config2.global_state_mapping['street']] = Street.RIVER
    global_state[config2.global_state_mapping['pot']] = 50
    winnings = game_over(global_state, config2, total_amount_invested)
    print('winnings',winnings)
    assert winnings[2]['result'] == -25
    assert winnings[6]['result'] == 25

def test_game_over_with_side_pots(config3):
    global_state = np.zeros(config3.global_state_shape)
    # set stacks for two players
    board_cards = [('A', 's'), ('K', 's'), ('Q', 's'), ('J', 's'), ('T', 's')]
    board = [readable_card_to_int(card) for card in board_cards]
    hand1 = [('A', 'h'), ('K', 'h'), ('Q', 'h'), ('J', 'h')]
    hand1 = [readable_card_to_int(card) for card in hand1]
    hand2 = [('A', 'c'), ('K', 'c'), ('Q', 'c'), ('J', 'c')]
    hand2 = [readable_card_to_int(card) for card in hand2]
    hand3 = [('A', 'c'), ('K', 'c'), ('9', 's'), ('8', 's')]
    hand3 = [readable_card_to_int(card) for card in hand3]
    # flatten hand1 and hand2
    hand1 = [item for sublist in hand1 for item in sublist]
    hand2 = [item for sublist in hand2 for item in sublist]
    hand3 = [item for sublist in hand3 for item in sublist]
    board = [item for sublist in board for item in sublist]
    hands = [hand1,hand2,hand3]
    stacks = [100, 100,0]
    for i,position in enumerate([1,2,6]):
        global_state[config3.global_state_mapping[f'player_{position}_stack']] = stacks[i]
        global_state[config3.global_state_mapping[f'player_{position}_active']] = 1
        global_state[config3.global_state_mapping[f'player_{position}_position']] = position
        global_state[config3.global_state_mapping[f'player_{position}_hand_range'][0]:config3.global_state_mapping[f'player_{position}_hand_range'][1]] = hands[i]
    total_amount_invested = {1: 50, 2: 50, 6:25}
    global_state[config3.global_state_mapping['board_range'][0]:config3.global_state_mapping['board_range'][1]] = board
    global_state[config3.global_state_mapping['street']] = Street.RIVER
    global_state[config3.global_state_mapping['pot']] = 125

    winnings = game_over(global_state, config3,total_amount_invested)
    print(winnings)
    assert winnings[1]['result'] == -25
    assert winnings[2]['result'] == -25
    assert winnings[6]['result'] == 50



from pokerrl_env.config import Config
from pokerrl_env.datatypes import StateActions
from pokerrl_env.utils import return_current_player
from pokerrl_env.view import player_view, human_readable_view
from pokerrl_env.transition import step_state,init_state

def get_action(action_mask,global_states,config):
    readable_actions = ['FOLD','CHECK','CALL']
    if global_states[-1,config.global_state_mapping['last_agro_action']] > StateActions.CALL:
        agro_action = 'RAISE'
    else:
        agro_action = 'BET'
    for betsize in config.betsizes:
        readable_actions.append(f'{agro_action} {betsize}')
    action_dict = {action:i for i,action in enumerate(readable_actions)}
    while True:
        possible_actions = [action for i,action in enumerate(readable_actions) if action_mask[i] > 0]
        try:
            print(list(range(len(possible_actions))))
            action_str = int(input(f'Enter the index of the desired action: {possible_actions}:'))
            return action_dict[possible_actions[action_str]]
        except ValueError:
            print('invalid action')

def play_game():
    config = Config(num_players=2)
    global_state,done,winnings,action_mask = init_state(config)
    while not done:
        player_idx = return_current_player(global_state)
        human_readable_view(global_state,player_idx, config,display=True)
        action = get_action(action_mask,global_state,config)
        global_state,done,winnings,action_mask = step_state(global_state, action, config)
    human_readable_view(global_state,global_state[-1,config.global_state_mapping['current_player']], config,display=True)
    print('winnings',winnings)

if __name__ == "__main__":
    play_game()


from typing import Dict, List, Tuple
import numpy as np
from pokerrl_env.config import Config
from pokerrl_env.datatypes import PLAYER_ORDER_BY_STREET, POSITION_TO_SEAT,RAISE,CALL,FOLD,BET,CHECK, ModelActions, StateActions,Street,Player,Positions
from pokerrl_env.cardlib import encode, hand_rank
from pokerrl_env.utils import is_next_player_the_aggressor, return_deck
import copy

def init_state(config: Config):
    deck = return_deck()
    
    state_SB = np.zeros(config.global_state_shape) 
    state_BB = np.zeros(config.global_state_shape)
    first_player = 'Dealer' if config.num_players == 2 else 'Small Blind'

    for state in (state_SB, state_BB):
        state[config.global_state_mapping["num_players"]] = config.num_players
        state[config.global_state_mapping["pot"]] = sum(config.blinds)

    if isinstance(config.stack_sizes, int):
        stack_sizes = [config.stack_sizes] * config.num_players
    else:
        stack_sizes = config.stack_sizes

    for i,position in enumerate(config.player_positions):
        hand = np.array([*deck.pop(), *deck.pop(),*deck.pop(), *deck.pop()],dtype=np.uint8)
        for state in (state_SB, state_BB):
            # Set player positions
            state[config.global_state_mapping[f"player_{position}_position"]] = position
            # Set player stack sizes
            state[config.global_state_mapping[f"player_{position}_stack"]] = stack_sizes[i]
            # Set player active status
            state[config.global_state_mapping[f"player_{position}_active"]] = 1
            # Set player hole cards
            state[config.global_state_mapping[f"player_{position}_hand_range"][0]:config.global_state_mapping[f"player_{position}_hand_range"][1]] = hand
    
    # board cards
    board_cards = np.array([*deck.pop(), *deck.pop(),*deck.pop(), *deck.pop(), *deck.pop()])

    # Small Blind action
    state_SB[config.global_state_mapping["current_player"]] = POSITION_TO_SEAT["Big Blind"]
    state_SB[config.global_state_mapping["previous_amount"]] = config.blinds[0]
    state_SB[config.global_state_mapping["previous_position"]] = POSITION_TO_SEAT[first_player]
    state_SB[config.global_state_mapping["previous_action"]] = 4  # Posting Small Blind is considered a raise
    state_SB[config.global_state_mapping["previous_bet_is_blind"]] = 1
    state_SB[config.global_state_mapping["last_agro_amount"]] = config.blinds[0]
    state_SB[config.global_state_mapping["last_agro_position"]] = POSITION_TO_SEAT[first_player]
    state_SB[config.global_state_mapping["last_agro_action"]] = StateActions.CALL + 1  # Posting Small Blind is considered a raise
    state_SB[config.global_state_mapping["last_agro_bet_is_blind"]] = 1
    state_SB[config.global_state_mapping["pot"]] = 0.5
    state_SB[config.global_state_mapping["next_player"]] = config.player_positions[1]
    state_SB[config.global_state_mapping["board_range"][0]:config.global_state_mapping["board_range"][1]] = board_cards
    state_SB[config.global_state_mapping["street"]] = Street.PREFLOP
    state_SB[config.global_state_mapping[f"player_{POSITION_TO_SEAT[first_player]}_stack"]] -= config.blinds[0]

    # Big Blind action
    state_BB[config.global_state_mapping["current_player"]] = config.player_positions[2 % config.num_players]
    state_BB[config.global_state_mapping["previous_amount"]] = config.blinds[1]
    state_BB[config.global_state_mapping["previous_position"]] = POSITION_TO_SEAT["Big Blind"]
    state_BB[config.global_state_mapping["previous_action"]] = StateActions.CALL + 1  # Posting Big Blind is considered a raise
    state_BB[config.global_state_mapping["previous_bet_is_blind"]] = 1
    state_BB[config.global_state_mapping["last_agro_amount"]] = config.blinds[1]
    state_BB[config.global_state_mapping["last_agro_position"]] = POSITION_TO_SEAT["Big Blind"]
    state_BB[config.global_state_mapping["last_agro_action"]] = 4  # Posting Small Blind is considered a raise
    state_BB[config.global_state_mapping["last_agro_bet_is_blind"]] = 1
    state_BB[config.global_state_mapping["pot"]] = 1.5
    state_BB[config.global_state_mapping["next_player"]] = config.player_positions[3 % config.num_players]
    state_BB[config.global_state_mapping["board_range"][0]:config.global_state_mapping["board_range"][1]] = board_cards
    state_BB[config.global_state_mapping["street"]] = Street.PREFLOP
    state_BB[config.global_state_mapping["player_2_stack"]] -= config.blinds[1]
    state_BB[config.global_state_mapping[f"player_{POSITION_TO_SEAT[first_player]}_stack"]] -= config.blinds[0]

    winnings = {position: 0 for position in config.player_positions}
    done = False
    player_totals = {position: 0 for position in config.player_positions}
    player_totals[POSITION_TO_SEAT[first_player]] = config.blinds[0]
    player_totals[POSITION_TO_SEAT["Big Blind"]] = config.blinds[1]
    return np.stack((state_SB, state_BB), axis=0),done,winnings,get_action_mask(state_BB,player_totals,config)

def clear_previous_action(global_state,config:Config):
    global_state[config.global_state_mapping[f'previous_action']] = 0
    global_state[config.global_state_mapping[f'previous_position']] = 0
    global_state[config.global_state_mapping[f'previous_amount']] = 0
    global_state[config.global_state_mapping[f'previous_bet_is_blind']] = 0

def clear_last_agro_action(global_state,config:Config):
    global_state[config.global_state_mapping[f'last_agro_action']] = 0
    global_state[config.global_state_mapping[f'last_agro_position']] = 0
    global_state[config.global_state_mapping[f'last_agro_amount']] = 0
    global_state[config.global_state_mapping[f'last_agro_bet_is_blind']] = 0

def players_finished(global_state,config:Config):
    num_active_players = 0
    for position in config.player_positions:
        if global_state[config.global_state_mapping[f'player_{position}_active']] == 1 and global_state[config.global_state_mapping[f'player_{position}_stack']] > 0:
            num_active_players += 1
            if num_active_players > 1:
                return False
    return True

def classify_action(action,player_street_total,player_stack,last_agro_amount,last_agro_action,pot,config:Config) -> Tuple[ModelActions, int]: 
    """ Returns action string and associated betsize """
    if action == ModelActions.FOLD:
        return FOLD, 0
    elif action == ModelActions.CHECK:
        return CHECK, 0
    elif action == ModelActions.CALL:
        return CALL, last_agro_amount - player_street_total
    elif action > ModelActions.CALL:
        # either bet or raise.
        return config.return_betsize(last_agro_action,last_agro_amount,config,action,pot,player_street_total,player_stack)
    else:
        raise ValueError(f"Invalid action: {action}")

def order_players_by_street(global_state:np.ndarray,config:Config):
    active_players = []
    for i in range(1,7):
        if global_state[config.global_state_mapping[f'player_{i}_stack']] > 0 and global_state[config.global_state_mapping[f'player_{i}_active']] == 1:
            active_players.append(Player(global_state[config.global_state_mapping[f'player_{i}_position']],global_state[config.global_state_mapping[f'player_{i}_stack']],global_state[config.global_state_mapping[f'player_{i}_active']]))
    player_ordering = PLAYER_ORDER_BY_STREET[int(global_state[config.global_state_mapping['street']])]
    active_players.sort(key=lambda x: player_ordering[x.position])
    return active_players

def get_pots(pot_players: List[Player], total_amount_invested: Dict[int, float]) -> List[Tuple[float, List[Player]]]:
    pots = []
    while pot_players:
        num_players = len(pot_players)
        min_invested = total_amount_invested[min(pot_players, key=lambda x: total_amount_invested[x.position]).position]
        pot = 0
        involved_players = []
        players_to_remove = []
        for player in pot_players:
            involved_players.append(player)
            pot += min_invested
            total_amount_invested[player.position] -= min_invested
            if total_amount_invested[player.position] == 0:
                players_to_remove.append(player)
        for player in players_to_remove:
            pot_players.remove(player)
        pots.append([pot, involved_players])
        if num_players == len(pot_players):
            break
    if sum(total_amount_invested.values()) > 0:
        pots[-1][0] += sum(total_amount_invested.values())
    return pots

def calculate_winnings(pots: List[Tuple[float, List[Player]]], players: List[Player]) -> Dict[int, float]:
    winnings = {}
    for p in players:
        winnings[p.position] = {
            'hand': p.hand if p.active else [],
            'hand_value': p.hand_value if p.active else 0,
            'result': -p.total_invested
        }
    for pot, involved_players in pots:
        min_hand_rank = min(involved_players, key=lambda x: x.hand_value).hand_value
        winners = [player for player in involved_players if player.hand_value == min_hand_rank]
        player_winnings = pot / len(winners)
        for winner in winners:
            winner.stack += player_winnings
            winnings[winner.position]['result'] += player_winnings
    return winnings

def game_over(global_state:np.ndarray,config:Config,total_amount_invested:float):
    # get all hand values
    board = global_state[config.global_state_mapping[f'board_range'][0]:config.global_state_mapping[f'board_range'][1]]
    board = [int(h) for h in board]
    en_board = [encode(board[i*2:(i*2)+2]) for i in range(0,len(board)//2)]
    players = []
    for position in config.player_positions:
        hand_start = config.global_state_mapping[f'player_{position}_hand_range'][0]
        hand_end = config.global_state_mapping[f'player_{position}_hand_range'][1]
        player_hand = [int(h) for h in global_state[hand_start:hand_end]]
        encoded_hand = [encode(player_hand[i*2:(i*2)+2]) for i in range(0,len(player_hand)//2)]
        hand_value = hand_rank(encoded_hand, en_board)
        players.append(Player(position=global_state[config.global_state_mapping[f'player_{position}_position']],
                                    stack=global_state[config.global_state_mapping[f'player_{position}_stack']],
                                    active=global_state[config.global_state_mapping[f'player_{position}_active']],
                                    hand=player_hand,
                                    hand_value=hand_value,
                                    total_invested=total_amount_invested[position]))
    
    pot_players = [p for p in players if p.active > 0]

    if len(pot_players) > 1 and global_state[config.global_state_mapping['street']] < 4:
        # accelerate street to river
        print('Accelerating street to river')
        global_state = create_next_state(global_state,config,street=4)
        print(global_state.shape)
    # Identify side pots and main pot
    pots = get_pots(pot_players, total_amount_invested)
    # Find winners and distribute the pots
    winnings = calculate_winnings(pots, players)
    return winnings

def increment_players(global_state:np.ndarray,active_players:list,current_player:int,config:Config):
    """ Skip players with stack 0. Which can happen if a player when allin but there are 2+ active players remaining """
    try:
        global_state[config.global_state_mapping['current_player']] = global_state[config.global_state_mapping['next_player']]
        non_zero_players = [p for p in active_players if p.stack > 0]
        player_idx = [p.position for p in non_zero_players].index(current_player)
        for player in non_zero_players:
            if player.position == current_player:
                next_player = non_zero_players[(player_idx + 2) % len(non_zero_players)]
        global_state[config.global_state_mapping[f'next_player']] = next_player.position
    except ValueError as e:
        print(active_players)
        print(current_player)
        raise e

def new_street_player_order(global_state:np.ndarray,config:Config):
    """ Skip players with stack 0. Which can happen if a player when allin but there are 2+ active players remaining """
    active_players = order_players_by_street(global_state,config)
    print('active_players',active_players)
    non_zero_players = [p for p in active_players if p.stack > 0]
    global_state[config.global_state_mapping[f'current_player']] = non_zero_players[0].position if len(non_zero_players) > 0 else 0
    global_state[config.global_state_mapping[f'next_player']] = non_zero_players[1].position if len(non_zero_players) > 1 else 0

def create_next_state(global_state:np.ndarray,config:Config,street:int):
    """ Creates a new global state by copying the current global state and clearing the previous action and last agro action. """
    new_global_state = np.copy(global_state)
    print('new_global_state',new_global_state.shape)
    new_global_state[config.global_state_mapping['street']] = street
    clear_previous_action(new_global_state,config)
    clear_last_agro_action(new_global_state,config)
    new_street_player_order(new_global_state,config)
    return np.stack((global_state,new_global_state))


def get_action_mask(global_state, player_amount_invested_per_street, config:Config):
    current_player_position = int(global_state[config.global_state_mapping["current_player"]])
    if current_player_position > 0:
        current_player_stack = global_state[config.global_state_mapping[f"player_{current_player_position}_stack"]]
        if current_player_stack > 0:
            pot = global_state[config.global_state_mapping[f"pot"]]
            current_player_investment = player_amount_invested_per_street[current_player_position]
            return config.return_action_mask(global_state,config,pot,current_player_investment,current_player_stack)
    return np.zeros(config.num_actions)

def return_investments(global_states,config:Config):
    player_amount_invested_per_street = {position:0 for position in config.player_positions}
    player_total_amount_invested = {position:0 for position in config.player_positions}
    current_street = 1
    print('number of states',global_states.shape[0])
    for global_state in global_states:
        previous_player = global_state[config.global_state_mapping['previous_position']]

        if global_state[config.global_state_mapping['street']] > current_street:
            # new street
            print('new street')
            player_amount_invested_per_street = {position:0 for position in config.player_positions}
            current_street = global_state[config.global_state_mapping['street']]
        else:
            print('same street')
            if global_state[config.global_state_mapping[f'previous_action']] > StateActions.CALL and not global_state[config.global_state_mapping[f'previous_bet_is_blind']]:
                # Special case for raise
                player_total_amount_invested[previous_player] += global_state[config.global_state_mapping[f'previous_amount']] - player_amount_invested_per_street[previous_player]
                player_amount_invested_per_street[previous_player] += global_state[config.global_state_mapping[f'previous_amount']] - player_amount_invested_per_street[previous_player]
            else:
                player_total_amount_invested[previous_player] += global_state[config.global_state_mapping[f'previous_amount']]
                player_amount_invested_per_street[previous_player] += global_state[config.global_state_mapping[f'previous_amount']]
        print('street',global_state[config.global_state_mapping[f'street']])
        print('previous_action',global_state[config.global_state_mapping[f'previous_action']])
        print('previous_amount',global_state[config.global_state_mapping[f'previous_amount']])
        print(player_amount_invested_per_street, player_total_amount_invested)
    return player_amount_invested_per_street, player_total_amount_invested

def step_state(global_states:np.ndarray, action:int, config:Config):
    """ Step the state forward by one action. Record the total amount invested by each player per street. """
    player_amount_invested_per_street, player_total_amount_invested = return_investments(global_states,config)
    
    # calculate next state
    global_state = np.copy(global_states[-1])
    active_players = order_players_by_street(global_state,config)
    current_player = int(global_state[config.global_state_mapping['current_player']])
    # Get action details
    action_category,betsize = classify_action(action, player_amount_invested_per_street[current_player], global_state[config.global_state_mapping[f'player_{current_player}_stack']], global_state[config.global_state_mapping['last_agro_amount']],global_state[config.global_state_mapping['last_agro_action']],global_state[config.global_state_mapping['pot']],config)
    player_total_amount_invested[current_player] += betsize
    global_state[config.global_state_mapping[f'pot']] += betsize
    if action_category == RAISE:
        global_state[config.global_state_mapping[f'player_{current_player}_stack']] -= betsize - player_amount_invested_per_street[current_player]
    else:
        global_state[config.global_state_mapping[f'player_{current_player}_stack']] -= betsize

    global_state[config.global_state_mapping[f'previous_action']] = config.convert_model_action_to_state(action)
    global_state[config.global_state_mapping[f'previous_position']] = current_player
    global_state[config.global_state_mapping[f'previous_amount']] = betsize
    global_state[config.global_state_mapping[f'previous_bet_is_blind']] = 0
    done = False
    winnings = {position:{'hand_value':0,'hand':[],'result':0} for position in config.player_positions}
    if action_category in [BET, RAISE]:
        # Subtract the amount already invested.
        global_state[config.global_state_mapping[f'pot']] -= player_amount_invested_per_street[current_player]
        # update last agro action
        global_state[config.global_state_mapping[f'last_agro_action']] = config.convert_model_action_to_state(action)
        global_state[config.global_state_mapping[f'last_agro_position']] = current_player
        global_state[config.global_state_mapping[f'last_agro_amount']] = betsize
        global_state[config.global_state_mapping[f'last_agro_bet_is_blind']] = 0
        # update next player
        increment_players(global_state,active_players,current_player,config)

    elif action_category == CHECK:
        if current_player == active_players[-1].position:
            # end of street
            if global_state[config.global_state_mapping[f'street']] == Street.RIVER:
                # end of game
                done = True
                winnings = game_over(global_state,config,player_total_amount_invested)
            else:
                # update street
                global_state = create_next_state(global_state,config,global_state[config.global_state_mapping['street']] + 1)
        else:
            increment_players(global_state,active_players,current_player,config)
    elif action_category == CALL:
        # Special case preflop blind situation.
        if global_state[config.global_state_mapping['street']] == Street.PREFLOP and \
            global_state[config.global_state_mapping['last_agro_amount']] == config.blinds[1] and \
            global_state[config.global_state_mapping['next_player']] == Positions.BIG_BLIND and \
            global_state[config.global_state_mapping[f'last_agro_position']] == Positions.BIG_BLIND:
            # bb can raise, or check
            increment_players(global_state,active_players,current_player,config)
        elif global_state[config.global_state_mapping['next_player']] == global_state[config.global_state_mapping['last_agro_position']] or len(active_players) == 1 or global_state[config.global_state_mapping['last_agro_position']] not in active_players and is_next_player_the_aggressor(active_players,current_player,global_state[config.global_state_mapping['last_agro_position']]):
            # end of round. last street or all players allin
            if global_state[config.global_state_mapping['street']] == Street.RIVER or players_finished(global_state, config):
                # end of game
                done = True
                winnings = game_over(global_state,config,player_total_amount_invested)
            else:
                # update street
                global_state = create_next_state(global_state,config,global_state[config.global_state_mapping['street']] + 1)
        else:
            increment_players(global_state,active_players,current_player,config)
    elif action_category == FOLD:
        global_state[config.global_state_mapping[f'player_{current_player}_active']] = 0
        # check for end of game
        if players_finished(global_state, config):
            # end game
            done = True
            winnings = game_over(global_state,config,player_total_amount_invested)
        elif global_state[config.global_state_mapping['street']] == Street.PREFLOP and \
            global_state[config.global_state_mapping['last_agro_amount']] == config.blinds[1] and \
            global_state[config.global_state_mapping['next_player']] == Positions.BIG_BLIND and \
            global_state[config.global_state_mapping[f'last_agro_position']] == Positions.BIG_BLIND:
            # bb can raise, or check
            increment_players(global_state,active_players,current_player,config)
        elif global_state[config.global_state_mapping['next_player']] == global_state[config.global_state_mapping['last_agro_position']]:
            # end of round
            if global_state[config.global_state_mapping['street']] == Street.RIVER:
                # end of game
                winnings = game_over(global_state,config,player_total_amount_invested)
                done = True
            else:
                # update street
                global_state = create_next_state(global_state,config,global_state[config.global_state_mapping['street']] + 1)
        else:
            increment_players(global_state,active_players,current_player,config)

    # To account for when the street updates and we output 2 states
    if len(global_state.shape) == 1:
        global_states = np.concatenate([global_states,global_state[None,:]],axis=0)
    else:
        print(global_states.shape,global_state.shape)
        global_states = np.concatenate([global_states,global_state])
    return global_states,done,winnings,get_action_mask(global_states[-1],player_amount_invested_per_street,config)